from pykechain.exceptions import _DeprecationMixin
from pykechain.models import ActivityReferencesProperty


class ActivityReferencesProperty2(ActivityReferencesProperty, _DeprecationMixin):
    """A virtual object representing a KE-chain Activity References property.

    .. versionadded:: 3.7
    """

    pass
