predefined_mimes = {
    "image/*": [
        "image/png",
        "image/gif",
        "image/jpeg",
        "image/webp",
        "image/tiff",
        "image/svg+xml",
        "image/bmp",
    ],
    ".xls": [
        "application/vnd.ms-excel",
        "application/x-excel",
        "application/x-msexcel",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    ],
    ".xlsx": [
        "application/vnd.ms-excel",
        "application/x-excel",
        "application/x-msexcel",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    ],
    "video/*": [
        "video/x-flv",
        "video/mp4",
        "application/x-mpegURL",
        "video/MP2T",
        "video/3gpp",
        "video/quicktime",
        "video/x-msvideo",
        "video/x-ms-wmv",
        "video/mpeg",
        "video/mp1s",
        "video/ogg",
        "video/webm",
    ],
    "model/*": [
        "model/3mf",
        "model/example",
        "model/gltf-binary",
        "model/gltf+json",
        "model/iges",
        "model/stl",
        "model/vnd.collada+xml",
        "model/vnd.dwf",
        "model/vnd.flatland.3dml",
        "model/vnd.gdl",
        "model/vnd.gs-gdl",
        "model/vnd.gtw",
        "model/vnd.moml+xml",
        "model/vnd.mts",
        "model/vnd.opengex",
        "model/vnd.usdz+zip",
        "model/vnd.valve.source.compiled-map",
        "model/vnd.vtu",
        "model/x3d-vrml",
        "model/x3d+fastinfoset",
        "model/x3d+xml",
    ],
}
